/*****************************************************************************
 * File:	utils.h
 *
 * Author:	Rhett "Jonzy" Jones
 *		jonzy@cc.utah.edu
 *
 * Date:	March 6, 1993
 *
 * Modifed:	None yet.
 *
 * Description:	Header file for use with "utils.c".
 *
 * Bugs:	No known bugs.
 *
 * Copyright:	Copyright 1993, University of Utah Computer Center.
 *		This source may be freely distributed as long as this copyright
 * 		notice remains intact, and is in no way used for any monetary
 * 		gain, by any institution, business, person, or persons.
 *
 ****************************************************************************/

char	*StrToLower(char *str);
long	GetLong(FILE *fp);
int	GetInt(FILE *fp);
char	*GetStr(FILE *fp,char *str,int limit);
int	Str2Int(char *s);
char	*Reverse(char *s, char *r);
short	StrRcmp(char *s1, char *s2);
int	DoSystemCall(char *format,...);
char	*MyStrTok(char *str,char delimeter);
char	*OnlyDigits(char *s);
char	*RemoveCRLF(char *line);
